/* Copyright 2006 by Mega Man. */
#include "stdio.h"
#include "cdvd.h"
#include "mass_rpc.h"
#include "libmc.h"
#include "libhdd.h"
#include "fileXio_rpc.h"
#include "sys/fcntl.h"

/* XXX: not working need to disable iop reset and directory listing function. */
//#define HOST_SUPPORT


#define MC_MAX 64
#define MAX_PARTITIONS 20

/* Put it into a union to save mem. */
typedef union {
	mcTable memCardEntry[MC_MAX];
	t_hddFilesystem partitions[MAX_PARTITIONS];
} fileEntries_t;

fileEntries_t mem;
static int hddFound = 0;
static char hddpath[256];
static char lastMounted[256];

int myStrlen(const char *name);
int myStrncmp(const char *cmp1, const char *cmp2, int n);

int myStrcmp(const char *cmp1, const char *cmp2)
{
	while ((*cmp1 != 0) && (*cmp2 != 0))
	{
		if (*cmp1 != *cmp2)
			return -1;
		cmp1++;
		cmp2++;
	}
	if (*cmp1 != *cmp2)
		return -1;
	else
		return 0;
}

int myStrncmp(const char *cmp1, const char *cmp2, int n)
{
	int i;

	i = 0;
	while ((*cmp1 != 0) && (*cmp2 != 0))
	{
		if (*cmp1 != *cmp2)
			return -1;
		cmp1++;
		cmp2++;
		i++;
		if (i >= n)
			return 0;
	}
	if (*cmp1 != *cmp2)
		return -1;
	else
		return 0;
}

int puts(const char *s)
{
	fioWrite(1, s, myStrlen(s));
	return 1;
}

int fakeCDVDInit(void)
{
	int ret;
	const char hddArgs[] = "-o" "\0" "4" "\0" "-n" "\0" "20";
	const char pfsArgs[] = "-m" "\0" "4" "\0" "-o" "\0" "10" "\0" "-n" "\0" "40";

#define loadModule(name, argv, argc) \
	{ \
		extern unsigned char name[]; \
		extern unsigned int size_##name; \
		ret = loadModuleBuffer(name, size_##name, argc, argv); \
		if (ret < 0) { \
			printf("Error " #name ".irx\n"); \
		} else { \
			printf(#name ".irx ok\n"); \
		} \
	}

	loadModule(usbd, NULL, 0);
	loadModule(usb_mass, NULL, 0);
	loadModule(poweroff, NULL, 0);
	loadModule(iomanX, NULL, 0);
	loadModule(fileXio, NULL, 0);
	loadModule(ps2dev9, NULL, 0);
	loadModule(ps2atad, NULL, 0);
	loadModule(ps2hdd, hddArgs, sizeof(hddArgs));
	loadModule(ps2fs, pfsArgs, sizeof(pfsArgs));

	ret = usb_mass_bindRpc();
	if (ret < 0 ) {
		printf("Error usb\n");
	}

	fileXioInit();
	if(hddCheckPresent() < 0)
	{
		printf("Errror hdd\n");
		hddFound = 0;
	}
	else
	{
		printf("Hdd\n");
		hddFound = 1;
	}
	lastMounted[0] = 0;

	return CDVD_Init();
}

int traceFioOpen(char *name, int mode)
{
	int n;
	printf("fioOpen(%s, %d)\n", name, mode);
	if (myStrncmp(name, "cdfs:/USB", 9) == 0)
	{
		name[3] = 'm';
		name[4] = 'a';
		name[5] = 's';
		name[6] = 's';
		name[7] = '0';
		name[8] = ':';
		printf("load (%s)\n", &name[3]);
		return fileXioOpen(&name[3], mode, 0);
	} else if (myStrncmp(name, "cdfs:/CDROM", 11) == 0)
	{
		name[6] = 'c';
		name[7] = 'd';
		name[8] = 'f';
		name[9] = 's';
		name[10] = ':';
		printf("load (%s)\n", &name[6]);
		return fileXioOpen(&name[6], mode, 0);
#ifdef HOST_SUPPORT
	} else if (myStrncmp(name, "cdfs:/HOST", 10) == 0)
	{
		name[5] = 'h';
		name[6] = 'o';
		name[7] = 's';
		name[8] = 't';
		name[9] = ':';
		printf("load (%s)\n", &name[5]);
		return fileXioOpen(&name[5], mode, 0);
#endif
	} else if (myStrncmp(name, "cdfs:/MC", 8) == 0)
	{
		name[6] = 'm';
		name[7] = 'c';
		name[9] = ':';
		printf("load (%s)\n", &name[6]);
		return fileXioOpen(&name[6], mode, 0);
	} else if (myStrncmp(name, "cdfs:pfs0:", 10) == 0)
	{
		printf("load (%s)\n", &name[5]);
		return fileXioOpen(&name[5], mode, 0);
	} else if (myStrncmp(name, "cdfs:/pfs0:", 11) == 0)
	{
		printf("load (%s)\n", &name[6]);
		return fileXioOpen(&name[6], mode, 0);
	} else
		return fileXioOpen(name, mode, 0);
}

int myStrlen(const char *name)
{
	int i;

	i = 0;
	while((name[i] != 0) && i < 2048)
	{
		i++;
	}
	return i;
}

void myStrcpy(char *dst, const char *src)
{
	do
	{
		*dst = *src;
		dst++;
		src++;
	} while(*src != 0);
	*dst = *src;
}

void myMemset(char *dst, char value, int count)
{
	int i;

	for(i = 0; i < count; i++)
		dst[i] = value;
}

extern int realGetDir(const char* pathname, const char* extensions, enum CDVD_getMode getMode, struct TocEntry tocEntry[], unsigned int req_entries, char* new_pathname);

void fixPath(char *new_pathname)
{
	int n;

	if (new_pathname != NULL)
	{
		int n;

		if (strcmp(new_pathname, "pfs0:/..") == 0) {
			strcpy(new_pathname, "/");
#ifdef DEBUG
			printf("path \"%s\"\n", new_pathname);
#endif
			return;
		}

		// Remove trailing "/"
		n = myStrlen(new_pathname);
		while (new_pathname[n-1] == '/')
		{
			new_pathname[n-1] = 0;
			n--;
			if (n == 0)
				break;
		}

		n = myStrlen(new_pathname);
		if ((new_pathname[n-1] == '.')
			&& (new_pathname[n-2] == '.')
			&& (new_pathname[n-3] == '/'))
		{
			n = n - 4;
			while (new_pathname[n] != '/')
			{
				n--;
			}
			new_pathname[n] = 0;
		}
#ifdef DEBUG
		printf("path \"%s\"\n", new_pathname);
#endif
	}
}

void addFakeEntry(struct TocEntry *tocEntry)
{
	// Fake entry
	tocEntry->fileLBA = 0;
	tocEntry->fileSize = 2048;
	tocEntry->fileProperties = 2;
	tocEntry->filename[0] = '.';
	tocEntry->filename[1] = '.';
	tocEntry->filename[2] = 0;
}

int traceGetDir(const char* pathname, const char* extensions, enum CDVD_getMode getMode, struct TocEntry tocEntry[], unsigned int req_entries, char* new_pathname)
{
	int ret;
	int i;
	int j;
	char *p;

	printf("getDir(%s, %s, %d, 0x%08x, %d, 0x%08x)\n", pathname, extensions, getMode, tocEntry, req_entries, new_pathname);
	if (new_pathname == NULL)
		new_pathname = pathname;
	else
		myStrcpy(new_pathname, pathname);
	fixPath(new_pathname);

	if (myStrncmp(new_pathname, "/CDROM", 6) == 0)
	{
		int counter = 0;
		myMemset(tocEntry, 0, sizeof(struct TocEntry) * req_entries);

		if (myStrcmp(new_pathname, "/CDROM") == 0)
		{
			addFakeEntry(&tocEntry[counter]);
			counter++;
			new_pathname[6] = '/';
			new_pathname[7] = 0;
		}
		counter += realGetDir(&new_pathname[6], extensions, getMode, &tocEntry[counter], req_entries - counter, &new_pathname[6]);
		if (counter == 0)
		{
			addFakeEntry(&tocEntry[counter]);
			counter++;
			CDVD_FlushCache();
		}
		return counter;
#ifdef HOST_SUPPORT
	} else if (myStrncmp(new_pathname, "/HOST", 5) == 0)
	{
		int counter = 0;

		if (myStrcmp(new_pathname, "/HOST") == 0)
		{
			// Fake entry
			addFakeEntry(&tocEntry[counter]);
			counter++;
		}

		// XXX: How to do this?
		return counter;
#endif
	} else if (myStrncmp(new_pathname, "/USB", 4) == 0)
	{
		int counter = 0;
		fat_dir_record record;

		myMemset(tocEntry, 0, sizeof(struct TocEntry) * req_entries);
	
		if (myStrcmp(new_pathname, "/USB") == 0)
		{
			// Fake entry
			addFakeEntry(&tocEntry[counter]);
			counter++;
			new_pathname[4] = '/';
			new_pathname[5] = 0;
		}

		//printf("usb_mass_getFirstDirentry(%s)\n", &new_pathname[4]);
		ret = usb_mass_getFirstDirentry(&new_pathname[4], &record);
		while (ret > 0 ) {
			if (record.attr & 0x10)
			{
				int l;
				if ((record.name[0] != '.')
					|| (record.name[1] != 0))
				{
					// Directory
					tocEntry[counter].fileLBA = 0;
					tocEntry[counter].fileSize = 2048;
					tocEntry[counter].fileProperties = 2;
					myStrcpy(tocEntry[counter].filename, record.name);
				}
				counter++;
			} else if ((record.attr & 0x08) == 0x08) {
				// Volume label
			} else {
				if (record.size > 1024)
				{
					tocEntry[counter].fileLBA = 0;
					tocEntry[counter].fileSize = record.size;
					tocEntry[counter].fileProperties = 0;
					myStrcpy(tocEntry[counter].filename, record.name);
					counter++;
				}
#if 0
				else
				{
					printf("File %s is to small.\n", record.name);
				}
#endif
			}

			if (counter >= req_entries)
			{
#if 0
				printf("Not enough memory for all entries\n");
#endif
				break;
			}
			ret = usb_mass_getNextDirentry(&record);
		}
		fixPath(new_pathname);
		if (counter == 0)
		{
			addFakeEntry(&tocEntry[counter]);
			counter++;
		}
		return counter;
	} else	if (myStrncmp(new_pathname, "/MC", 3) == 0)
	{
		int counter = 0;
		int i;
		myMemset(tocEntry, 0, sizeof(struct TocEntry) * req_entries);

		if ((myStrcmp(new_pathname, "/MC0") == 0)
			|| (myStrcmp(new_pathname, "/MC1") == 0))
		{
			// Fake entry
			addFakeEntry(&tocEntry[counter]);
			counter++;
			new_pathname[4] = '/';
			new_pathname[5] = '*';
			new_pathname[6] = 0;
		}
		else
		{
			/* Add "*" to path. */
			i = myStrlen(new_pathname);
			if (new_pathname[i-1] != '/')
			{
				new_pathname[i] = '/';
				i++;
			}
			new_pathname[i] = '*';
			i++;
			new_pathname[i] = 0;
		}
#ifdef DEBUG
		printf("mcGetDir %s %d\n", &new_pathname[4], new_pathname[3] - '0');
#endif
		mcGetDir(new_pathname[3] - '0', 0, &new_pathname[5], 0, MC_MAX, mem.memCardEntry);
		mcSync(0, NULL, &ret);
#ifdef DEBUG
		printf("mcGetDir() ret %d\n", ret);
#endif
		i = 0;
		while((ret > i) && (counter < req_entries))
		{
			if ((mem.memCardEntry[i].name[0] != '.')
				|| (mem.memCardEntry[i].name[1] != 0))
			{
				if(mem.memCardEntry[i].attrFile & MC_ATTR_SUBDIR)
				{
					tocEntry[counter].fileProperties = 2;
				}
				else
				{
					tocEntry[counter].fileProperties = 0;
				}
				tocEntry[counter].fileLBA = 0;
				tocEntry[counter].fileSize = mem.memCardEntry[i].fileSizeByte;
				myStrcpy(tocEntry[counter].filename, mem.memCardEntry[i].name);
				counter++;
			}
			i++;
		}
		/* Remove '*' from String */
		i = myStrlen(new_pathname);
		new_pathname[i - 1] = 0;
		fixPath(new_pathname);
		if (counter == 0)
		{
			addFakeEntry(&tocEntry[counter]);
			counter++;
		}
		return counter;
	} else	if (myStrncmp(new_pathname, "/hdd0:", 6) == 0)
	{
		int rv;
		int counter = 0;
		int len;
		int dirFd;
		iox_dirent_t thisDir;
		int needMount = 1;

		addFakeEntry(&tocEntry[counter]);
		counter++;

#ifdef DEBUG
		printf("Mount %s\n", &new_pathname[1]);
#endif

		if (lastMounted[0] != 0) {
			if (strcmp(lastMounted, &new_pathname[1]) == 0) {
				needMount = 0;
			} else {
				fileXioUmount("pfs0:");
				needMount = 1;
			}
		} else {
			needMount = 1;
		}
		if (needMount) {
			rv = fileXioMount("pfs0:", &new_pathname[1], FIO_MT_RDONLY);
			if(rv != 0)
			{
				lastMounted[0] = 0;
#ifdef DEBUG
				printf("mount failed\n");
#endif
				return counter;
			}
			strcpy(lastMounted, &new_pathname[1]);
		}
		len = myStrlen(new_pathname);
		if (new_pathname[len-1] != '/')
		{
			new_pathname[len] = '/';
			len++;
			new_pathname[len] = 0;
		}
		/* Find first slash -> beginning of path. */
		for (i = 6; i < len; i++)
		{
			if (new_pathname[i] == '/')
				break;
		}

		myStrcpy(hddpath, "pfs0:");
		myStrcpy(&hddpath[5], &new_pathname[i]);
#ifdef DEBUG
		printf("Opening hdd path \"%s\"\n", hddpath);
#endif
		dirFd = fileXioDopen(hddpath);
		if(dirFd >= 0)
		{
			rv = fileXioDread(dirFd, &thisDir);
			while((rv > 0) && (counter < req_entries))
			{
#ifdef DEBUG
				printf("File: %s attr 0x%x\n", thisDir.name, thisDir.stat.attr);
#endif
				if(strcmp(thisDir.name, ".") && strcmp(thisDir.name, ".."))
				{
					tocEntry[counter].fileLBA = 0;
					if (FIO_SO_ISDIR(thisDir.stat.attr)) {
						tocEntry[counter].fileSize = 2048;
						tocEntry[counter].fileProperties = 2;
					} else {
						tocEntry[counter].fileSize = thisDir.stat.size;
						tocEntry[counter].fileProperties = 0;
					}

					myStrcpy(tocEntry[counter].filename, thisDir.name);
					counter++;
				}
				rv = fileXioDread(dirFd, &thisDir);
			}
			fileXioDclose(dirFd);
		}
		else
		{
#ifdef DEBUG
			printf("failed hdd open.\n");
#endif
		}
		strcpy(new_pathname, hddpath);
		fixPath(new_pathname);
		return counter;
	} else	if (myStrncmp(new_pathname, "pfs0:", 4) == 0)
	{
		int rv;
		int counter = 0;
		int len;
		int dirFd;
		iox_dirent_t thisDir;

		addFakeEntry(&tocEntry[counter]);
		counter++;

		len = myStrlen(new_pathname);
		if (new_pathname[len-1] != '/')
		{
			new_pathname[len] = '/';
			len++;
			new_pathname[len] = 0;
		}
#ifdef DEBUG
		printf("Opening hdd path \"%s\"\n", new_pathname);
#endif
		dirFd = fileXioDopen(new_pathname);
		if(dirFd >= 0)
		{
			rv = fileXioDread(dirFd, &thisDir);
			while((rv > 0) && (counter < req_entries))
			{
#ifdef DEBUG
				printf("File: %s\n", thisDir.name);
#endif
				if(strcmp(thisDir.name, ".") && strcmp(thisDir.name, ".."))
				{
					tocEntry[counter].fileLBA = 0;
					if (FIO_SO_ISDIR(thisDir.stat.attr)) {
						tocEntry[counter].fileSize = 2048;
						tocEntry[counter].fileProperties = 2;
					} else {
						tocEntry[counter].fileSize = thisDir.stat.size;
						tocEntry[counter].fileProperties = 0;
					}
					myStrcpy(tocEntry[counter].filename, thisDir.name);
					counter++;
				}
				rv = fileXioDread(dirFd, &thisDir);
			}
			fileXioDclose(dirFd);
		}
		else
		{
#ifdef DEBUG
			printf("failed hdd open.\n");
#endif
		}
		fixPath(new_pathname);
		return counter;

	} else {
		int counter = 0;
		tocEntry[counter].fileLBA = 0x0002b8d2;
		tocEntry[counter].fileSize = 2048;
		tocEntry[counter].fileProperties = 2;
		myStrcpy(tocEntry[counter].filename, "CDROM");
		counter++;

		tocEntry[counter].fileLBA = 0;
		tocEntry[counter].fileSize = 2048;
		tocEntry[counter].fileProperties = 2;
		myStrcpy(tocEntry[counter].filename, "USB");
		counter++;

#ifdef HOST_SUPPORT
		tocEntry[counter].fileLBA = 0;
		tocEntry[counter].fileSize = 2048;
		tocEntry[counter].fileProperties = 2;
		myStrcpy(tocEntry[counter].filename, "HOST");
		counter++;
#endif
		tocEntry[counter].fileLBA = 0;
		tocEntry[counter].fileSize = 2048;
		tocEntry[counter].fileProperties = 2;
		myStrcpy(tocEntry[counter].filename, "MC0");
		counter++;

		tocEntry[counter].fileLBA = 0;
		tocEntry[counter].fileSize = 2048;
		tocEntry[counter].fileProperties = 2;
		myStrcpy(tocEntry[counter].filename, "MC1");
		counter++;

		if (hddFound) {
			int n;

			if (lastMounted[0] != 0) {
				fileXioUmount("pfs0:");
				lastMounted[0] = 0;
			}
			n = hddGetFilesystemList(mem.partitions, 20);
#ifdef DEBUG
			printf("Found %i partitions\n", n);
#endif

			for (i = 0; i < n; i++) {
				tocEntry[counter].fileLBA = 0;
				tocEntry[counter].fileSize = 2048;
				tocEntry[counter].fileProperties = 2;
				myStrcpy(tocEntry[counter].filename, mem.partitions[i].filename);
				counter++;
			}
		}

		if (new_pathname != NULL)
		{
			new_pathname[0] = 0;
		}
		return counter;
	}
}

