/* Copyright 2006 by Mega Man. */
#include <stdio.h>
#include <lzo1x.h>
#include <malloc.h>

typedef unsigned int u32;

typedef struct {
	u32 compressedSize;
	u32 originalSize;
	u32 zeroByteSize;		// specifies the bytes which are to be zeroed
	u32 virtualAddr;
	u32 fileSize;
} SjCRUNCH2_SectionHeader;

typedef struct {
	u32 entryAddr;
	u32 numSections;
} SjCRUNCH2_Header;

#define assert(x) \
	do { \
		if (!(x)) { \
			fprintf(stderr, "Error: Asssert failed in line %d\n", __LINE__); \
			return -4; \
		} \
	} while(0)

int main(int argc, char *argv[])
{
	FILE *fin;
	FILE *fout;
	SjCRUNCH2_Header header;
	SjCRUNCH2_SectionHeader section;
	int rv;
	int i;
	unsigned char *compressed = NULL;
	unsigned char *uncompressed = NULL;
	unsigned char *wrk = NULL;

	if (argc != 3)
	{
		fprintf(stderr, "sjuncrunch [crunched elf] [uncrunched elf output]\n\n");
		fprintf(stderr, "Decompress a sjcrunched file.\n");
		fprintf(stderr, "Result is a binary file.\n");
		fprintf(stderr, "Currently only intented to use with one section,\n");
		fprintf(stderr, "because all sections will be written to one file.\n");
		return -1;
	}

	fin = fopen(argv[1], "rb");
	if (fin == NULL)
	{
		fprintf(stderr, "Error: Can't open file \"%s\".\n", argv[1]);
		return -2;
	}
	rv = fseek(fin, 0x2f00, SEEK_SET);
	assert(rv == 0);
	rv = fread(&header, sizeof(header), 1, fin);
	assert(rv == 1);
	printf("entry 0x%08x\n", header.entryAddr);
	printf("sections %d\n", header.numSections);

	rv = lzo_init();
	assert(rv == LZO_E_OK);

	fout = fopen(argv[2], "wb");
	if (fout == NULL)
	{
		fprintf(stderr, "Error: Can't open file \"%s\".\n", argv[2]);
		return -5;
	}

	for (i = 0; i < header.numSections; i++)
	{
		unsigned int compressedSize;
		unsigned int uncompressedSize;
		unsigned int newSize;
		unsigned int total;

		rv = fread(&section, sizeof(section), 1, fin);
		assert(rv == 1);

		printf("compressedSize %d\n", section.compressedSize);
		printf("originalSize %d\n", section.originalSize);
		printf("zeroByteSize %d\n", section.zeroByteSize);
		printf("virtualAddr 0x%08x\n", section.virtualAddr);
		printf("fileSize %d\n", section.fileSize);

		total = 0;

		compressed = malloc(section.compressedSize);
		assert(compressed != NULL);
		uncompressed = malloc(section.originalSize);
		assert(uncompressed != NULL);
		//wrk = malloc(LZO1X_999_MEM_COMPRESS);
		//assert(wrk != NULL);

		while(total < section.originalSize)
		{
			rv = fread(&uncompressedSize, 4, 1, fin);
			assert(rv == 1);
			rv = fread(&compressedSize, 4, 1, fin);
			assert(rv == 1);
			printf("Compressed size %d\n", compressedSize);
			printf("Uncompressed size %d\n", uncompressedSize);
			//assert(size == section.compressedSize);
			rv = fread(compressed, 1, compressedSize, fin);
			assert(rv == compressedSize);

			if (uncompressedSize != compressedSize)
			{
				newSize = 0;
				rv = lzo1x_decompress(compressed, compressedSize,
					&uncompressed[total], &newSize,
					NULL);
				printf("newSize %d rv %d\n", newSize, rv);
				assert(rv == LZO_E_OK);
			}
			else
			{
				/* Block was not compressed. */
				memcpy(&uncompressed[total], compressed, uncompressedSize);
				newSize = uncompressedSize;
			}
			assert(uncompressedSize == newSize);
			total += newSize;
			printf("+ Total %d\n", total);
		}
		assert(total == section.originalSize);
		rv = fwrite(uncompressed, 1, section.originalSize, fout);
		assert(rv == section.originalSize);
		free(compressed);
		compressed = NULL;
		free(uncompressed);
		uncompressed = NULL;
		free(wrk);
		wrk = NULL;
	}

	fclose(fin);
	fclose(fout);
	return 0;
}

